using CombinatorialSpaces
using ComponentArrays
using OrdinaryDiffEq
using GeometryBasics
using LinearAlgebra
using Base.Iterators
using Catlab
using MLStyle
import Catlab.Programs.GenerateJuliaPrograms: compile


const gensim_in_place_stub = Symbol("GenSim-M")

abstract type AbstractCall end

struct UnaryCall <: AbstractCall 
    operator
    equality
    input
    output
end

# TODO: Add back support for contract operators
Base.Expr(c::UnaryCall) = begin
    operator = c.operator
    #= if isa(operator, AbstractArray)
        operator = Expr(:call, :∘, reverse(operator)...)
    end =#
    if(c.equality == :.=)
        if(operator == add_inplace_stub(:⋆₁⁻¹)) # Since inverse hodge Geo is a solver
            Expr(:call, c.operator, c.output, c.input)
        else
            Expr(:call, :mul!, c.output, operator, c.input)
        end
    else
        Expr(c.equality, c.output, Expr(:call, operator, c.input))
    end
end
                
struct BinaryCall <: AbstractCall 
    operator
    equality
    input1
    input2
    output
end

# TODO: After getting rid of AppCirc2, do we need this check?
Base.Expr(c::BinaryCall) = begin
    #= if isa(c.operator, AbstractArray)
        operator = :(compose($(c.operator)))
    end =#

    # These operators can be done in-place
    if(c.equality == :.= && get_stub(c.operator) == gensim_in_place_stub)
        return Expr(:call, c.operator, c.output, c.input1, c.input2)
    end
    return Expr(c.equality, c.output, Expr(:call, c.operator, c.input1, c.input2))
end

struct VarargsCall <: AbstractCall 
    operator
    equality
    inputs
    output
end

Base.Expr(c::VarargsCall) = begin
    #= if isa(c.operator, AbstractArray)
        operator = :(compose($(c.operator)))
    end =#
    return Expr(c.equality, c.output, Expr(:call, c.operator, c.inputs...))
end

struct AllocVecCall <: AbstractCall 
    name
    form
    dimension
    T
end

struct AllocVecCallError <: Exception
    c::AllocVecCall
end

# TODO: There are likely better ways of dispatching on dimension instead of
# storing it inside an AllocVecCall.
Base.Expr(c::AllocVecCall) = begin
    resolved_form = @match (c.form, c.dimension) begin
        (:Form0, 2) => :V
        (:Form1, 2) => :E
        (:Form2, 2) => :Tri
        (:DualForm0, 2) => :Tri
        (:DualForm1, 2) => :E
        (:DualForm2, 2) => :V

        (:Form0, 1) => :V
        (:Form1, 1) => :E
        (:DualForm0, 1) => :E
        (:DualForm1, 1) => :V
        _ => throw(AllocVecCallError(c))
    end

    :($(Symbol(:__,c.name)) = Decapodes.FixedSizeDiffCache(Vector{$(c.T)}(undef, nparts(mesh, $(QuoteNode(resolved_form))))))
end

#= function get_form_number(d::SummationDecapode, var_id::Int)
    type = d[var_id, :type]
    if(type == :Form0)
        return 0
    elseif(type == :Form1)
        return 1
    elseif(type == :Form2)
        return 2
    end
    return -1
end

# WARNING: This may not work if names are not unique, use above instead
function get_form_number(d::SummationDecapode, var_name::Symbol)
    var_id = first(incident(d, var_name, :name))
    return get_form_number(d, var_id)
end =#

function is_form(d::SummationDecapode, var_id::Int)
    type = d[var_id, :type]
    return (type == :Form0 || type == :Form1 || type == :Form2 || 
        type == :DualForm0 || type == :DualForm1 || type == :DualForm2)
end

is_form(d::SummationDecapode, var_name::Symbol) = is_form(d, first(incident(d, var_name, :name)))

is_literal(d::SummationDecapode, var_id::Int) = (d[var_id, :type] == :Literal)
is_literal(d::SummationDecapode, var_name::Symbol) = is_literal(d, first(incident(d, var_name, :name)))

is_infer(d::SummationDecapode, var_id::Int) = (d[var_id, :type] == :infer)
is_infer(d::SummationDecapode, var_name::Symbol) = is_infer(d, first(incident(d, var_name, :name)))

add_stub(stub_name::Symbol, var_name::Symbol) = return Symbol("$(stub_name)_$(var_name)")

function get_stub(var_name::Symbol)
    var_str = String(var_name)
    idx = findfirst("_", var_str)
    if(isnothing(idx) || first(idx) == 1)
        return nothing
    end
    return Symbol(var_str[begin:first(idx) - 1])
end

add_inplace_stub(var_name::Symbol) = add_stub(gensim_in_place_stub, var_name)

# This will be the function and matrix generation
function compile_env(d::AbstractNamedDecapode, dec_matrices::Vector{Symbol}, con_dec_operators::Set{Symbol})
    assumed_ops = Set([:+, :*, :-, :/, :.+, :.*, :.-, :./])
    defined_ops = Set()

    defs = quote end

    for op in dec_matrices
        if(op in defined_ops )
            continue
        end

        quote_op = QuoteNode(op)
        mat_op = add_stub(gensim_in_place_stub, op)
        # Could turn this into a special call
        def = :(($mat_op, $op) = default_dec_matrix_generate(mesh, $quote_op, hodge))
        push!(defs.args, def)

        push!(defined_ops, op)
    end

    for op in d[:op1]
      if op == DerivOp 
        continue
      end

      if(op in con_dec_operators || op in defined_ops)
          continue
      end
  
      ops = QuoteNode(op)
      def = :($op = operators(mesh, $ops))
        
      push!(defs.args, def)
  
      push!(defined_ops, op)
    end
    for op in d[:op2]
      if op in assumed_ops || op in defined_ops
        continue
      end
      ops = QuoteNode(op)
      def = :($op = operators(mesh, $ops))
      push!(defs.args, def)
  
      push!(defined_ops, op)
    end
    return defs
  end

function compile_var(alloc_vectors::Vector{AllocVecCall})
    return quote $(Expr.(alloc_vectors)...) end
end

# This is the block of parameter setting inside f
# TODO: Pass this an extra type parameter that sets the size of the Floats
get_vars_code(d::AbstractNamedDecapode, vars::Vector{Symbol}) = get_vars_code(d, vars, Float64)

function get_vars_code(d::AbstractNamedDecapode, vars::Vector{Symbol}, ::Type{stateeltype}) where stateeltype
    stmts = map(vars) do s
        ssymbl = QuoteNode(s)
        if all(d[incident(d, s, :name) , :type] .== :Constant)
        #if only(d[incident(d, s, :name) , :type]) == :Constant
            :($s = p.$s)
        elseif all(d[incident(d, s, :name) , :type] .== :Parameter)
            :($s = (p.$s)(t))
        elseif all(d[incident(d, s, :name) , :type] .== :Literal)
            # Literals don't need assignments, because they are literals, but we stored them as Symbols.
            # #TODO: we should fix that upstream so that we don't need this.
            :($s = $(parse(stateeltype, String(s))))
        else
            # TODO: If names are not unique, then the type is assumed to be a
            # form for all of the vars sharing a same name.
            :($s = u.$s)
        end
    end
    return quote $(stmts...) end
end

# This is the setting of the du vector at the end of f
function set_tanvars_code(d::AbstractNamedDecapode)
    tanvars = [(d[e, [:src,:name]], d[e, [:tgt,:name]]) for e in incident(d, :∂ₜ, :op1)]
    stmts = map(tanvars) do (s,t)
        ssymb = QuoteNode(s)
        :(getproperty(du, $ssymb) .= $t)
    end
    return stmts
end

function compile(d::SummationDecapode, inputs::Vector, alloc_vectors::Vector{AllocVecCall}, optimizable_dec_operators::Set{Symbol}; dimension=2, stateeltype=Float64)
    # Get the Vars of the inputs (probably state Vars).
    visited_Var = falses(nparts(d, :Var))

    # input_numbers = incident(d, inputs, :name)
    input_numbers = reduce(vcat, incident(d, inputs, :name))
    # visited_Var[collect(flatten(input_numbers))] .= true
    visited_Var[input_numbers] .= true
    visited_Var[incident(d, :Literal, :type)] .= true

    visited_1 = falses(nparts(d, :Op1))
    visited_2 = falses(nparts(d, :Op2))
    visited_Σ = falses(nparts(d, :Σ))

    promote_arithmetic_map = Dict(:(+) => :.+, :(-) => :.-, :(*) => :.*, :(/) => :./, :(=) => :.=,
                                  :.+ => :.+, :.- => :.-, :.* => :.*, :./ => :./, :.= => :.=)

    # FIXME: this is a quadratic implementation of topological_sort inlined in here.
    op_order = []
    for _ in 1:(nparts(d, :Op1) + nparts(d,:Op2) + nparts(d, :Σ))
        for op in parts(d, :Op1)
            s = d[op, :src]
            if !visited_1[op] && visited_Var[s]
                # skip the derivative edges
                operator = d[op, :op1]
                t = d[op, :tgt]
                visited_1[op] = true
                if operator == DerivOp
                    continue
                end

                equality = :(=)

                sname = d[s, :name]
                tname = d[t, :name]

                # TODO: Check to see if this is a DEC operator
                if(operator in optimizable_dec_operators)
                    # push!(dec_matrices, operator)

                    if(is_form(d, t))
                        equality = promote_arithmetic_map[equality]
                        operator = add_stub(gensim_in_place_stub, operator)

                        push!(alloc_vectors, AllocVecCall(tname, d[t, :type], dimension, stateeltype))
                    end
                end


                visited_Var[t] = true
                c = UnaryCall(operator, equality, sname, tname)
                push!(op_order, c)
            end
        end

        for op in parts(d, :Op2)
            arg1 = d[op, :proj1]
            arg2 = d[op, :proj2]
            if !visited_2[op] && visited_Var[arg1] && visited_Var[arg2]
                r = d[op, :res]
                a1name = d[arg1, :name]
                a2name = d[arg2, :name]
                rname  = d[r, :name]

                operator = d[op, :op2]
                equality = :(=)

                # This is meant for wedge products
                #= if(operator in optimizable_dec_operators)
                    push!(dec_matrices, operator)
                end =#

                # TODO: Check to make sure that this logic never breaks
                if(is_form(d, r))
                    if(operator == :(+) || operator == :(-) || operator == :.+ || operator == :.-)
                        operator = promote_arithmetic_map[operator]
                        equality = promote_arithmetic_map[equality]
                        push!(alloc_vectors, AllocVecCall(rname, d[r, :type], dimension, stateeltype))
                    
                    # TODO: Do we want to support the ability of a user to use the backslash operator?
                    elseif(operator == :(*) || operator == :(/) || operator == :.* || operator == :./)
                        # WARNING: This part may break if we add more compiler types that have different 
                        # operations for basic and broadcast modes, e.g. matrix multiplication vs broadcast
                        if(!is_infer(d, arg1) && !is_infer(d, arg2))
                            operator = promote_arithmetic_map[operator]
                            equality = promote_arithmetic_map[equality]
                            push!(alloc_vectors, AllocVecCall(rname, d[r, :type], dimension, stateeltype))
                        end
                    elseif(operator in optimizable_dec_operators)
                        operator = add_stub(gensim_in_place_stub, operator)
                        equality = promote_arithmetic_map[equality]
                        push!(alloc_vectors, AllocVecCall(rname, d[r, :type], dimension, stateeltype))
                    end
                end

                if(operator == :(*))
                    operator = promote_arithmetic_map[operator]
                end
                if(operator == :(-))
                    operator = promote_arithmetic_map[operator]
                end

                visited_2[op] = true
                visited_Var[r] = true
                c = BinaryCall(operator, equality, a1name, a2name, rname)
                push!(op_order, c)
            end
        end

        for op in parts(d, :Σ)
            args = subpart(d, incident(d, op, :summation), :summand)
            if !visited_Σ[op] && all(visited_Var[args])
                r = d[op, :sum]
                argnames = d[args, :name]
                rname  = d[r, :name]

                operator = :(+)
                equality = :(=)

                # If result is a known form, broadcast addition
                # TODO: Also need to tell handler to prealloc
                if(is_form(d, r))
                    operator = promote_arithmetic_map[operator]
                    equality = promote_arithmetic_map[equality]
                    push!(alloc_vectors, AllocVecCall(rname, d[r, :type], dimension, stateeltype))
                end

                operator = :(.+)

                visited_Σ[op] = true
                visited_Var[r] = true
                c = VarargsCall(operator, equality, argnames, rname)
                push!(op_order, c)
            end
        end
    end

    cache_exprs = map(alloc_vectors) do vec
        :($(vec.name) = (Decapodes.get_tmp($(Symbol(:__,vec.name)), u)))
    end

    eq_exprs = map(Expr, op_order)
    vcat(cache_exprs,eq_exprs)
end
  
# TODO: Add more specific types later for optimization
function resolve_types_compiler!(d::SummationDecapode)
    d[:type] = map(d[:type]) do x
        if(x == :Constant || x == :Parameter)
            return :infer
        end
        return x
    end
end

function replace_negation_with_multiply!(d::SummationDecapode, input_vars)
    found_negation = false
    rem_negations = []
    neg1var = 0

    for (i, op) in enumerate(d[:op1])
        if(op == :(-) || op == :neg)
            if(!found_negation)
                neg1var = add_part!(d, :Var, type = :Literal, name = Symbol("-1.0"))
                push!(input_vars, Symbol("-1.0"))
                found_negation = true
            end
            push!(rem_negations, i)
            add_part!(d, :Op2, proj1 = neg1var, proj2 = d[i, :src], res = d[i, :tgt], op2 = :.*)
        end
        
    end
    rem_parts!(d, :Op1, rem_negations)
end

function replace_names_compiler!(d::SummationDecapode)
    dec_op1 = Pair{Symbol, Any}[]
    dec_op2 = Pair{Symbol, Symbol}[(:∧₀₀ => :.*)]
    replace_names!(d, dec_op1, dec_op2)
end

function infer_overload_compiler!(d::SummationDecapode, dimension::Int)
    if(dimension == 1)
        infer_types!(d, op1_inf_rules_1D, op2_inf_rules_1D)
        resolve_overloads!(d, op1_res_rules_1D, op2_res_rules_1D)
    elseif(dimension == 2)
        infer_types!(d, op1_inf_rules_2D, op2_inf_rules_2D)
        resolve_overloads!(d, op1_res_rules_2D, op2_res_rules_2D)
    end
end

function init_dec_matrices!(d::SummationDecapode, dec_matrices::Vector{Symbol}, optimizable_dec_operators::Set{Symbol})

    for op1_name in d[:op1]
        if(op1_name ∈ optimizable_dec_operators)
            push!(dec_matrices, op1_name)
        end
    end

    for op2_name in d[:op2]
        if(op2_name ∈ optimizable_dec_operators)
            push!(dec_matrices, op2_name)
        end
    end
end

function link_contract_operators(d::SummationDecapode, con_dec_operators::Set{Symbol})

    contract_defs = quote end

    compute_to_name = Dict()
    curr_id = 1

    for op1_id in parts(d, :Op1)
        op1_name = d[op1_id, :op1]
        if isa(op1_name, AbstractArray)
            computation = reverse!(map(x -> add_inplace_stub(x), op1_name))
            compute_key = join(computation, " * ")

            computation_name = get(compute_to_name, compute_key, :Error)
            if(computation_name == :Error)
                computation_name = add_stub(Symbol("GenSim-ConMat"), Symbol(curr_id))
                get!(compute_to_name, compute_key, computation_name)
                push!(con_dec_operators, computation_name)

                expr_line = Expr(Symbol("="), add_inplace_stub(computation_name), Expr(:call, :*, computation...))
                push!(contract_defs.args, expr_line)

                expr_line = Expr(Symbol("="), computation_name, Expr(Symbol("->"), :x, Expr(:call, :*, add_inplace_stub(computation_name), :x)))
                push!(contract_defs.args, expr_line)

                curr_id += 1
            end

            d[op1_id, :op1] = computation_name
        end
    end

    contract_defs
end

# TODO: Will want to eventually support contracted operations
function gensim(user_d::AbstractNamedDecapode, input_vars; dimension::Int=2,
                stateeltype = Float64)
    # TODO: May want to move this after infer_types if we let users
    # set their own inference rules
    recognize_types(user_d)

    # Makes copy
    d′ = expand_operators(user_d)
    #d′ = average_rewrite(d′)

    replace_negation_with_multiply!(d′, input_vars)

    dec_matrices = Vector{Symbol}();
    alloc_vectors = Vector{AllocVecCall}();

    vars = get_vars_code(d′, input_vars, stateeltype)
    tars = set_tanvars_code(d′)
    
    # We need to run this after we grab the constants and parameters out
    resolve_types_compiler!(d′)
    infer_overload_compiler!(d′, dimension)

    # This should probably be followed by an expand_operators
    replace_names_compiler!(d′)
    open_operators!(d′, dimension = dimension)
    infer_overload_compiler!(d′, dimension)

    # This will generate all of the fundemental DEC operators present
    optimizable_dec_operators = Set([:⋆₀, :⋆₁, :⋆₂, :⋆₀⁻¹, :⋆₂⁻¹,
                                    :d₀, :d₁, :dual_d₀, :d̃₀, :dual_d₁, :d̃₁])
    extra_dec_operators = Set([:⋆₁⁻¹, :∧₀₁, :∧₁₀, :∧₁₁, :∧₀₂, :∧₂₀])

    init_dec_matrices!(d′, dec_matrices, union(optimizable_dec_operators, extra_dec_operators))

    # This contracts matrices together into a single matrix
    contracted_dec_operators = Set{Symbol}();
    contract_operators!(d′, allowable_ops = optimizable_dec_operators)
    cont_defs = link_contract_operators(d′, contracted_dec_operators)

    union!(optimizable_dec_operators, contracted_dec_operators, extra_dec_operators)

    # Compilation of the simulation
    equations = compile(d′, input_vars, alloc_vectors, optimizable_dec_operators, dimension=dimension, stateeltype=stateeltype)

    func_defs = compile_env(d′, dec_matrices, contracted_dec_operators)
    vect_defs = compile_var(alloc_vectors)

    quote
        function simulate(mesh, operators, hodge=GeometricHodge())
            $func_defs
            $cont_defs
            $vect_defs
            f(du, u, p, t) = begin
                $vars
                $(equations...)
                $(tars...)
            end;
        end
    end
end

gensim(c::Collage; dimension::Int=2) = 
    gensim(collate(c); dimension=dimension)

"""    function gensim(d::AbstractNamedDecapode; dimension::Int=2)

Generate a simulation function from the given Decapode. The returned function can then be combined with a mesh and a function describing function mappings to return a simulator to be passed to `solve`.
"""
gensim(d::AbstractNamedDecapode; dimension::Int=2, stateeltype = Float64) = gensim(d,
    vcat(collect(infer_state_names(d)), d[incident(d, :Literal, :type), :name]), dimension=dimension, stateeltype=stateeltype)

evalsim(d::AbstractNamedDecapode; dimension::Int=2, stateeltype = Float64) = eval(gensim(d, dimension=dimension, stateeltype=stateeltype))
evalsim(d::AbstractNamedDecapode, input_vars; dimension::Int=2, stateeltype = Float64) = eval(gensim(d, input_vars, dimension=dimension, stateeltype=stateeltype))
    
"""
    function find_unreachable_tvars(d)

Determine if the given Decapode can be compiled to an explicit time-stepping
simulation. Use the simple check that one can traverse the Decapode starting
from the state variables and reach all TVars (ignoring ∂ₜ edges).
"""
function find_unreachable_tvars(d)
  states = infer_states(d)
  visited = falses(nparts(d, :Var))
  s = Stack{Int64}()
  foreach(v -> push!(s, v), states)
  while true
    curr_var = pop!(s)
    visited[curr_var] = true
    op1s = incident(d, curr_var, :src)
    filter!(x -> d[x, :op1] == :∂ₜ, op1s)
    for op1 in op1s
      (!visited[d[op1, :tgt]]) && push!(s, d[op1, :tgt])
    end
  end
  TVars = d[:incl]
end

function closest_point(p1, p2, dims)
    p_res = collect(p2)
    for i in 1:length(dims)
        if dims[i] != Inf
            p = p1[i] - p2[i]
            f, n = modf(p / dims[i])
            p_res[i] += dims[i] * n
            if abs(f) > 0.5
                p_res[i] += sign(f) * dims[i]
            end
        end
    end
    Point3{Float64}(p_res...)
end

function flat_op(s::AbstractDeltaDualComplex2D, X::AbstractVector; dims=[Inf, Inf, Inf])
  # XXX: Creating this lookup table shouldn't be necessary. Of course, we could
  # index `tri_center` but that shouldn't be necessary either. Rather, we should
  # loop over incident triangles instead of the elementary duals, which just
  # happens to be inconvenient.
  tri_map = Dict{Int,Int}(triangle_center(s,t) => t for t in triangles(s))

  map(edges(s)) do e
    p = closest_point(point(s, tgt(s,e)), point(s, src(s,e)), dims)
    e_vec = (point(s, tgt(s,e)) - p) * sign(1,s,e)
    dual_edges = elementary_duals(1,s,e)
    dual_lengths = dual_volume(1, s, dual_edges)
    mapreduce(+, dual_edges, dual_lengths) do dual_e, dual_length
      X_vec = X[tri_map[s[dual_e, :D_∂v0]]]
      dual_length * dot(X_vec, e_vec)
    end / sum(dual_lengths)
  end
end
